@@ -6,14 +6,16 @@ mod cli {
         fs::{self, File},
         io::Write,
         path::{Path, PathBuf},
+        time::Duration,
     };
 
+    use anyhow::anyhow;
     use assert_cmd::Command;
     use assert_json_diff::assert_json_include;
     use http::StatusCode;
     use lychee_lib::{InputSource, ResponseBody};
     use predicates::{
-        prelude::predicate,
+        prelude::{predicate, PredicateBooleanExt},
         str::{contains, is_empty},
     };
     use pretty_assertions::assert_eq;
@@ -86,7 +88,7 @@ mod cli {
         errors: usize,
         cached: usize,
         success_map: HashMap<InputSource, HashSet<ResponseBody>>,
-        fail_map: HashMap<InputSource, HashSet<ResponseBody>>,
+        error_map: HashMap<InputSource, HashSet<ResponseBody>>,
         suggestion_map: HashMap<InputSource, HashSet<ResponseBody>>,
         excluded_map: HashMap<InputSource, HashSet<ResponseBody>>,
     }
@@ -111,6 +113,105 @@ mod cli {
         }};
     }
 
+    /// Test that the default report output format (compact) and mode (color)
+    /// prints the failed URLs as well as their status codes on error. Make
+    /// sure that the status code only occurs once.
+    #[test]
+    fn test_compact_output_format_contains_status() -> Result<()> {
+        let test_path = fixtures_path().join("TEST_INVALID_URLS.html");
+
+        let mut cmd = main_command();
+        cmd.arg("--format")
+            .arg("compact")
+            .arg("--mode")
+            .arg("color")
+            .arg(test_path)
+            .env("FORCE_COLOR", "1")
+            .assert()
+            .failure()
+            .code(2);
+
+        let output = cmd.output()?;
+
+        // Check that the output contains the status code (once) and the URL
+        let output_str = String::from_utf8_lossy(&output.stdout);
+
+        // The expected output is as follows:
+        // "Find details below."
+        // [EMPTY LINE]
+        // [path/to/file]:
+        //      [400] https://httpbin.org/status/404
+        //      [500] https://httpbin.org/status/500
+        //      [502] https://httpbin.org/status/502
+        // (the order of the URLs may vary)
+
+        // Check that the output contains the file path
+        assert!(output_str.contains("TEST_INVALID_URLS.html"));
+
+        let re = Regex::new(r"\s{5}\[\d{3}\] https://httpbin\.org/status/\d{3}").unwrap();
+        let matches: Vec<&str> = re.find_iter(&output_str).map(|m| m.as_str()).collect();
+
+        // Check that the status code occurs only once
+        assert_eq!(matches.len(), 3);
+
+        Ok(())
+    }
+
+    /// Test JSON output format
+    #[tokio::test]
+    async fn test_json_output() -> Result<()> {
+        // Server that returns a bunch of 200 OK responses
+        let mock_server_ok = mock_server!(StatusCode::OK);
+        let mut cmd = main_command();
+        cmd.arg("--format")
+            .arg("json")
+            .arg("-vv")
+            .arg("--no-progress")
+            .arg("-")
+            .write_stdin(mock_server_ok.uri())
+            .assert()
+            .success();
+        let output = cmd.output().unwrap();
+        let output_json = serde_json::from_slice::<Value>(&output.stdout)?;
+
+        // Check that the output is valid JSON
+        assert!(output_json.is_object());
+        // Check that the output contains the expected keys
+        assert!(output_json.get("detailed_stats").is_some());
+        assert!(output_json.get("success_map").is_some());
+        assert!(output_json.get("error_map").is_some());
+        assert!(output_json.get("excluded_map").is_some());
+
+        // Check the success map
+        let success_map = output_json["success_map"].as_object().unwrap();
+        assert_eq!(success_map.len(), 1);
+
+        // Get the actual URL from the mock server for comparison
+        let mock_url = mock_server_ok.uri();
+
+        // Create the expected success map structure
+        let expected_success_map = serde_json::json!({
+            "stdin": [
+                {
+                    "status": {
+                        "code": 200,
+                        "text": "200 OK"
+                    },
+                    "url": format!("{mock_url}/"),
+                }
+            ]
+        });
+
+        // Compare the actual success map with the expected one
+        assert_eq!(
+            success_map,
+            expected_success_map.as_object().unwrap(),
+            "Success map doesn't match expected structure"
+        );
+
+        Ok(())
+    }
+
     /// JSON-formatted output should always be valid JSON.
     /// Additional hints and error messages should be printed to `stderr`.
     /// See https://github.com/lycheeverse/lychee/issues/1355
@@ -150,12 +251,13 @@ mod cli {
         // Check that the output is valid JSON
         assert!(serde_json::from_slice::<Value>(&output.stdout).is_ok());
 
-        // Parse site error status from the fail_map
+        // Parse site error status from the error_map
         let output_json = serde_json::from_slice::<Value>(&output.stdout).unwrap();
-        let site_error_status = &output_json["fail_map"][&test_path.to_str().unwrap()][0]["status"];
+        let site_error_status =
+            &output_json["error_map"][&test_path.to_str().unwrap()][0]["status"];
 
         assert_eq!(
-            "error sending request for url (https://expired.badssl.com/)",
+            "error sending request for url (https://expired.badssl.com/) Maybe a certificate error?",
             site_error_status["details"]
         );
         Ok(())
@@ -231,6 +333,17 @@ mod cli {
         Ok(())
     }
 
+    #[test]
+    fn test_stylesheet_misinterpreted_as_email() -> Result<()> {
+        test_json_output!(
+            "TEST_STYLESHEET_LINK.md",
+            MockResponseStats {
+                total: 0,
+                ..MockResponseStats::default()
+            }
+        )
+    }
+
     /// Test that a GitHub link can be checked without specifying the token.
     #[test]
     fn test_check_github_no_token() -> Result<()> {
@@ -267,17 +380,53 @@ mod cli {
     #[test]
     fn test_resolve_paths() {
         let mut cmd = main_command();
-        let offline_dir = fixtures_path().join("offline");
+        let dir = fixtures_path().join("resolve_paths");
 
         cmd.arg("--offline")
             .arg("--base")
-            .arg(&offline_dir)
-            .arg(offline_dir.join("index.html"))
+            .arg(&dir)
+            .arg(dir.join("index.html"))
             .env_clear()
             .assert()
             .success()
-            .stdout(contains("4 Total"))
-            .stdout(contains("4 OK"));
+            .stdout(contains("3 Total"))
+            .stdout(contains("3 OK"));
+    }
+
+    #[test]
+    fn test_resolve_paths_from_root_dir() {
+        let mut cmd = main_command();
+        let dir = fixtures_path().join("resolve_paths_from_root_dir");
+
+        cmd.arg("--offline")
+            .arg("--include-fragments")
+            .arg("--root-dir")
+            .arg(&dir)
+            .arg(dir.join("nested").join("index.html"))
+            .env_clear()
+            .assert()
+            .failure()
+            .stdout(contains("7 Total"))
+            .stdout(contains("5 OK"))
+            .stdout(contains("2 Errors"));
+    }
+
+    #[test]
+    fn test_resolve_paths_from_root_dir_and_base_url() {
+        let mut cmd = main_command();
+        let dir = fixtures_path();
+
+        cmd.arg("--offline")
+            .arg("--root-dir")
+            .arg("/resolve_paths")
+            .arg("--base")
+            .arg(&dir)
+            .arg(dir.join("resolve_paths").join("index.html"))
+            .env_clear()
+            .assert()
+            .success()
+            .stdout(contains("3 Total"))
+            .stdout(contains("3 OK"));
     }
 
     #[test]
@@ -413,7 +562,7 @@ mod cli {
             .failure()
             .code(2)
             .stdout(contains(
-                "[404] https://github.com/mre/idiomatic-rust-doesnt-exist-man"
+                "[404] https://github.com/mre/idiomatic-rust-doesnt-exist-man | Network error: Not Found"
             ))
             .stderr(contains(
                 "There were issues with GitHub URLs. You could try setting a GitHub token and running lychee again.",
@@ -834,19 +983,103 @@ mod cli {
         let base_path = fixtures_path().join("cache");
         let cache_file = base_path.join(LYCHEE_CACHE_FILE);
 
+        // Ensure clean state
+        if cache_file.exists() {
+            println!("Removing cache file before test: {:?}", cache_file);
+            fs::remove_file(&cache_file)?;
+            tokio::time::sleep(Duration::from_millis(500)).await;
+        }
+
+        // Setup mock servers
+        let mock_server_ok = mock_server!(StatusCode::OK);
+        let mock_server_err = mock_server!(StatusCode::NOT_FOUND);
+        let mock_server_exclude = mock_server!(StatusCode::OK);
+
+        // Create test file
+        let dir = tempfile::tempdir()?;
+        let file_path = dir.path().join("c.md");
+        let mut file = File::create(&file_path)?;
+        writeln!(file, "{}", mock_server_ok.uri().as_str())?;
+        writeln!(file, "{}", mock_server_err.uri().as_str())?;
+        writeln!(file, "{}", mock_server_exclude.uri().as_str())?;
+        file.sync_all()?;
+
+        // Create and run command
+        let mut cmd = main_command();
+        cmd.current_dir(&base_path)
+            .arg(&file_path)
+            .arg("--verbose")
+            .arg("--no-progress")
+            .arg("--cache")
+            .arg("--exclude")
+            .arg(mock_server_exclude.uri());
+
+        // Note: Don't check output.status.success() since we expect
+        // a non-zero exit code (2) when lychee finds broken links
+        let _output = cmd.output()?;
+
+        // Wait for cache file to be written
+        for _ in 0..20 {
+            if cache_file.exists() {
+                break;
+            }
+            tokio::time::sleep(Duration::from_millis(100)).await;
+        }
+
+        // Check cache contents
+        let data = fs::read_to_string(&cache_file)?;
+        println!("Cache file contents: {}", data);
+
+        assert!(
+            data.contains(&format!("{}/,200", mock_server_ok.uri())),
+            "Missing OK entry in cache"
+        );
+        assert!(
+            data.contains(&format!("{}/,404", mock_server_err.uri())),
+            "Missing error entry in cache"
+        );
+
+        // Run again to verify cache behavior
+        cmd.assert()
+            .stderr(contains(format!(
+                "[200] {}/ | OK (cached)\n",
+                mock_server_ok.uri()
+            )))
+            .stderr(contains(format!(
+                "[404] {}/ | Error (cached)\n",
+                mock_server_err.uri()
+            )));
+
+        // Clean up
+        fs::remove_file(&cache_file).map_err(|e| {
+            anyhow::anyhow!(
+                "Failed to remove cache file: {:?}, error: {}",
+                cache_file,
+                e
+            )
+        })?;
+
+        Ok(())
+    }
+
+    #[tokio::test]
+    async fn test_lycheecache_exclude_custom_status_codes() -> Result<()> {
+        let base_path = fixtures_path().join("cache");
+        let cache_file = base_path.join(LYCHEE_CACHE_FILE);
+
         // Unconditionally remove cache file if it exists
         let _ = fs::remove_file(&cache_file);
 
         let mock_server_ok = mock_server!(StatusCode::OK);
-        let mock_server_err = mock_server!(StatusCode::NOT_FOUND);
-        let mock_server_exclude = mock_server!(StatusCode::OK);
+        let mock_server_no_content = mock_server!(StatusCode::NO_CONTENT);
+        let mock_server_too_many_requests = mock_server!(StatusCode::TOO_MANY_REQUESTS);
 
         let dir = tempfile::tempdir()?;
         let mut file = File::create(dir.path().join("c.md"))?;
 
         writeln!(file, "{}", mock_server_ok.uri().as_str())?;
-        writeln!(file, "{}", mock_server_err.uri().as_str())?;
-        writeln!(file, "{}", mock_server_exclude.uri().as_str())?;
+        writeln!(file, "{}", mock_server_no_content.uri().as_str())?;
+        writeln!(file, "{}", mock_server_too_many_requests.uri().as_str())?;
 
         let mut cmd = main_command();
         let test_cmd = cmd
@@ -855,43 +1088,40 @@ mod cli {
             .arg("--verbose")
             .arg("--no-progress")
             .arg("--cache")
-            .arg("--exclude")
-            .arg(mock_server_exclude.uri());
+            .arg("--cache-exclude-status")
+            .arg("204,429");
 
         assert!(
             !cache_file.exists(),
             "cache file should not exist before this test"
         );
 
-        // run first without cache to generate the cache file
+        // Run first without cache to generate the cache file
         test_cmd
             .assert()
             .stderr(contains(format!("[200] {}/\n", mock_server_ok.uri())))
             .stderr(contains(format!(
-                "[404] {}/ | Failed: Network error: Not Found\n",
-                mock_server_err.uri()
-            )));
-
-        // check content of cache file
-        let data = fs::read_to_string(&cache_file)?;
-        assert!(data.contains(&format!("{}/,200", mock_server_ok.uri())));
-        assert!(data.contains(&format!("{}/,404", mock_server_err.uri())));
-
-        // run again to verify cache behavior
-        test_cmd
-            .assert()
-            .stderr(contains(format!(
-                "[200] {}/ | Cached: OK (cached)\n",
-                mock_server_ok.uri()
+                "[204] {}/ | 204 No Content: No Content\n",
+                mock_server_no_content.uri()
             )))
             .stderr(contains(format!(
-                "[404] {}/ | Cached: Error (cached)\n",
-                mock_server_err.uri()
+                "[429] {}/ | Network error: Too Many Requests\n",
+                mock_server_too_many_requests.uri()
             )));
 
-        // clear the cache file
-        fs::remove_file(&cache_file)?;
+        // Check content of cache file
+        let data = fs::read_to_string(&cache_file)?;
 
+        if data.is_empty() {
+            println!("Cache file is empty!");
+        }
+
+        assert!(data.contains(&format!("{}/,200", mock_server_ok.uri())));
+        assert!(!data.contains(&format!("{}/,204", mock_server_no_content.uri())));
+        assert!(!data.contains(&format!("{}/,429", mock_server_too_many_requests.uri())));
+
+        // Unconditionally remove the cache file
+        let _ = fs::remove_file(&cache_file);
         Ok(())
     }
 
@@ -933,11 +1163,11 @@ mod cli {
             .failure()
             .code(2)
             .stdout(contains(format!(
-                "[418] {}/ | Failed: Network error: I\'m a teapot",
+                "[418] {}/ | Network error: I\'m a teapot",
                 mock_server_teapot.uri()
             )))
             .stdout(contains(format!(
-                "[500] {}/ | Failed: Network error: Internal Server Error",
+                "[500] {}/ | Network error: Internal Server Error",
                 mock_server_server_error.uri()
             )));
 
@@ -956,11 +1186,11 @@ mod cli {
             .assert()
             .success()
             .stderr(contains(format!(
-                "[418] {}/ | Cached: OK (cached)",
+                "[418] {}/ | OK (cached)",
                 mock_server_teapot.uri()
             )))
             .stderr(contains(format!(
-                "[500] {}/ | Cached: OK (cached)",
+                "[500] {}/ | OK (cached)",
                 mock_server_server_error.uri()
             )));
 
@@ -996,7 +1226,7 @@ mod cli {
             .stderr(contains(format!(
                 "[IGNORED] {unsupported_url} | Unsupported: Error creating request client"
             )))
-            .stderr(contains(format!("[EXCLUDED] {excluded_url} | Excluded\n")));
+            .stderr(contains(format!("[EXCLUDED] {excluded_url}\n")));
 
         // The cache file should be empty, because the only checked URL is
         // unsupported and we don't want to cache that. It might be supported in
@@ -1157,8 +1387,9 @@ mod cli {
         Ok(())
     }
 
-    /// If base-dir is not set, don't throw an error in case we encounter
-    /// an absolute local link within a file (e.g. `/about`).
+    /// If `base-dir` is not set, don't throw an error in case we encounter
+    /// an absolute local link (e.g. `/about`) within a file.
+    /// Instead, simply ignore the link.
     #[test]
     fn test_ignore_absolute_local_links_without_base() -> Result<()> {
         let mut cmd = main_command();
@@ -1350,9 +1581,7 @@ mod cli {
             .arg("./NOT-A-REAL-TEST-FIXTURE.md")
             .assert()
             .failure()
-            .stderr(contains(
-                "Cannot find local file ./NOT-A-REAL-TEST-FIXTURE.md",
-            ));
+            .stderr(contains("Invalid file path: ./NOT-A-REAL-TEST-FIXTURE.md"));
 
         Ok(())
     }
@@ -1375,6 +1604,8 @@ mod cli {
     #[test]
     #[ignore = "Skipping test because it is flaky"]
     fn test_suggests_url_alternatives() -> Result<()> {
+        let re = Regex::new(r"http://web\.archive\.org/web/.*google\.com/jobs\.html").unwrap();
+
         for _ in 0..3 {
             // This can be flaky. Try up to 3 times
             let mut cmd = main_command();
@@ -1390,7 +1621,6 @@ mod cli {
             // We're looking for a suggestion that
             // - starts with http://web.archive.org/web/
             // - ends with google.com/jobs.html
-            let re = Regex::new(r"http://web\.archive\.org/web/.*google\.com/jobs\.html").unwrap();
             if re.is_match(&String::from_utf8_lossy(&output.stdout)) {
                 // Test passed
                 return Ok(());
@@ -1470,7 +1700,6 @@ mod cli {
     async fn test_cookie_jar() -> Result<()> {
         // Create a random cookie jar file
         let cookie_jar = NamedTempFile::new()?;
-
         let mut cmd = main_command();
         cmd.arg("--cookie-jar")
             .arg(cookie_jar.path().to_str().unwrap())
@@ -1483,15 +1712,13 @@ mod cli {
 
         // check that the cookie jar file contains the expected cookies
         let file = std::fs::File::open(cookie_jar.path()).map(std::io::BufReader::new)?;
-        let cookie_store = reqwest_cookie_store::CookieStore::load_json(file).unwrap();
+        let cookie_store = cookie_store::serde::json::load(file)
+            .map_err(|e| anyhow!("Failed to load cookie jar: {e}"))?;
         let all_cookies = cookie_store.iter_any().collect::<Vec<_>>();
-
         assert!(!all_cookies.is_empty());
         assert!(all_cookies.iter().all(|c| c.domain() == Some("google.com")));
-
         Ok(())
     }
-
     #[test]
     fn test_dump_inputs_glob_md() -> Result<()> {
         let pattern = fixtures_path().join("**/*.md");
@@ -1526,6 +1753,24 @@ mod cli {
         Ok(())
     }
 
+    #[test]
+    fn test_dump_inputs_glob_exclude_path() -> Result<()> {
+        let pattern = fixtures_path().join("**/*");
+
+        let mut cmd = main_command();
+        cmd.arg("--dump-inputs")
+            .arg(pattern)
+            .arg("--exclude-path")
+            .arg(fixtures_path().join("dump_inputs/subfolder"))
+            .assert()
+            .success()
+            .stdout(contains("fixtures/dump_inputs/subfolder/test.html").not())
+            .stdout(contains("fixtures/dump_inputs/subfolder/file2.md").not())
+            .stdout(contains("fixtures/dump_inputs/subfolder").not());
+
+        Ok(())
+    }
+
     #[test]
     fn test_dump_inputs_url() -> Result<()> {
         let mut cmd = main_command();
@@ -1590,10 +1835,12 @@ mod cli {
             .stderr(contains(
                 "fixtures/fragments/file1.md#kebab-case-fragment-1",
             ))
-            .stdout(contains("15 Total"))
-            .stdout(contains("12 OK"))
-            // 3 failures because of missing fragments
-            .stdout(contains("3 Errors"));
+            .stderr(contains("fixtures/fragments/file.html#top"))
+            .stderr(contains("fixtures/fragments/file2.md#top"))
+            .stdout(contains("25 Total"))
+            .stdout(contains("21 OK"))
+            // 4 failures because of missing fragments
+            .stdout(contains("4 Errors"));
     }
 
     #[test]
@@ -1608,4 +1855,92 @@ mod cli {
             .success()
             .stdout(contains("0 Errors"));
     }
+
+    /// Test relative paths
+    ///
+    /// Imagine a web server hosting a site with the following structure:
+    /// root
+    /// └── test
+    ///     ├── index.html
+    ///     └── next.html
+    ///
+    /// where `root/test/index.html` contains `<a href="next.html">next</a>`
+    /// When checking the link in `root/test/index.html` we should be able to
+    /// resolve the relative path to `root/test/next.html`
+    ///
+    /// Note that the relative path is not resolved to the root of the server
+    /// but relative to the file that contains the link.
+    #[tokio::test]
+    async fn test_resolve_relative_paths_in_subfolder() -> Result<()> {
+        let mock_server = wiremock::MockServer::start().await;
+
+        let body = r#"<a href="next.html">next</a>"#;
+        wiremock::Mock::given(wiremock::matchers::method("GET"))
+            .and(wiremock::matchers::path("/test/index.html"))
+            .respond_with(wiremock::ResponseTemplate::new(200).set_body_string(body))
+            .mount(&mock_server)
+            .await;
+
+        wiremock::Mock::given(wiremock::matchers::method("GET"))
+            .and(wiremock::matchers::path("/test/next.html"))
+            .respond_with(wiremock::ResponseTemplate::new(200))
+            .mount(&mock_server)
+            .await;
+
+        let mut cmd = main_command();
+        cmd.arg("--verbose")
+            .arg(format!("{}/test/index.html", mock_server.uri()))
+            .assert()
+            .success()
+            .stdout(contains("1 Total"))
+            .stdout(contains("0 Errors"));
+
+        Ok(())
+    }
+
+    #[tokio::test]
+    async fn test_json_format_in_config() -> Result<()> {
+        let mock_server = mock_server!(StatusCode::OK);
+        let config = fixtures_path().join("configs").join("format.toml");
+        let mut cmd = main_command();
+        cmd.arg("--config")
+            .arg(config)
+            .arg("-")
+            .write_stdin(mock_server.uri())
+            .env_clear()
+            .assert()
+            .success();
+
+        // Check that the output is in JSON format
+        let output = cmd.output().unwrap();
+        let output = std::str::from_utf8(&output.stdout).unwrap();
+        let json: serde_json::Value = serde_json::from_str(output)?;
+        assert_eq!(json["total"], 1);
+
+        Ok(())
+    }
+
+    #[tokio::test]
+    async fn test_retry() -> Result<()> {
+        let mock_server = wiremock::MockServer::start().await;
+
+        wiremock::Mock::given(wiremock::matchers::method("GET"))
+            .respond_with(ResponseTemplate::new(429))
+            .up_to_n_times(1)
+            .mount(&mock_server)
+            .await;
+
+        wiremock::Mock::given(wiremock::matchers::method("GET"))
+            .respond_with(ResponseTemplate::new(200))
+            .mount(&mock_server)
+            .await;
+
+        let mut cmd = main_command();
+        cmd.arg("-")
+            .write_stdin(mock_server.uri())
+            .assert()
+            .success();
+
+        Ok(())
+    }
 }
